// Copyright (c) Microsoft Corporation. All rights reserved.

using System;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Security.Permissions;
using System.Runtime.Serialization;

namespace Microsoft.Tools.CodeRunner
{
	/// <summary>
	/// A valid command line argument 
	/// </summary>
	public class CommandLineArgument
	{
		#region Public Instance Constructors

		/// <summary>
		/// Constructs a new command line argument.
		/// </summary>
		/// <param name="attribute">The attribute that defines the argument</param>
		/// <param name="propertyInfo">The arguments property information</param>
		public CommandLineArgument(CommandLineArgumentAttribute attribute, PropertyInfo propertyInfo)
		{
			// Set these two things first
			this.attribute = attribute;
			this.propertyInfo = propertyInfo;

			if (this.IsArray)
			{
			    // Get an element type if the argument is an array
				valueType = propertyInfo.PropertyType.GetElementType();
				
				if (valueType == typeof(object))
				{
					throw new NotSupportedException(String.Format(CultureInfo.InvariantCulture, 
						CommandLineParserResources.PropertyIsNotAStronglyTypedArray, propertyInfo.Name));
				}
			}
			else if (this.IsCollection)
			{
				// Locate Add method with 1 parameter that is not System.Object
				foreach (MethodInfo method in propertyInfo.PropertyType.GetMethods(BindingFlags.Public | BindingFlags.Instance))
				{
					if (method.Name == "Add" && method.GetParameters().Length == 1)
					{
						ParameterInfo parameter = method.GetParameters()[0];

						if (parameter.ParameterType != typeof(object))
						{
						    addMethod = method;
						    valueType = parameter.ParameterType;
						}
					}
					
					if (method.Name == "GetEnumerator" && method.GetParameters().Length == 0 && typeof(IEnumerator).IsAssignableFrom(method.ReturnType))
					{
					    getEnumeratorMethod = method;
					}
					
					if (addMethod != null && getEnumeratorMethod != null)
					    break;
				}

                // If we didn't find an appropriate Add method we can't write to the collection
                if (addMethod == null)
                {
                    throw new NotSupportedException(String.Format(CultureInfo.InvariantCulture,
                        CommandLineParserResources.PropertyHasNoAddMethodWithNonObjectParameter, propertyInfo.Name));
                }

                // The collection derives from ICollection, so it would be really surprising if it didn't implement GetEnumerator
                Debug.Assert(getEnumeratorMethod != null);
            }
			else
			{
			    // The argument is not an array or collection, so it's safe to take the property type as the value type
			    this.valueType = propertyInfo.PropertyType;
			}

			this.values = new ArrayList(AllowMultiple ? 4 : 1);
		}

		#endregion

		#region Public Instance Properties

		/// <summary>
		/// Gets the underlying <see cref="Type" /> of the argument.
		/// </summary>
		/// <remarks>
		/// If the <see cref="Type" /> of the argument is a collection type,
		/// this property will returns the underlying type of the collection.
		/// </remarks>
		public Type ValueType
		{
			get { return valueType; }
		}

		/// <summary>
		/// Gets the name of the argument.
		/// </summary>
		/// <value>The long name of the argument.</value>
		public string Name
		{
			get { return attribute.Name; }
		}

		/// <summary>
		/// Gets the short name of the argument.
		/// </summary>
		public string ShortName
		{
			get { return attribute.ShortName; }
		}

		/// <summary>
		/// Gets the description of the argument.
		/// </summary>
		public string Description
		{
			get { return attribute.Description; }
		}

		/// <summary>
		/// Gets the hint for the argument value.
		/// </summary>
		public string ValueHint
		{
			get { return attribute.ValueHint; }
		}

		/// <summary>
		/// Gets a value indicating whether an instance of this argument is set.
		/// </summary>
		public bool HasArgumentValue
		{
			get { return Count > 0; }
		}

		/// <summary>
		/// Gets a count of the number of instances of this argument.
		/// </summary>
		public int Count
		{
			get
			{
				if (values != null)
					return values.Count;
				else
					return 0;
			}
		}

		/// <summary>
		/// Gets a value indicating whether the argument can be specified multiple times.
		/// </summary>
		/// <value>
		/// <see langword="true" /> if the argument may be specified multiple 
		/// times; otherwise, <see langword="false" />.
		/// </value>
		public bool AllowMultiple
		{
			get { return valueType != null && (IsCollection || IsArray); }
		}

		/// <summary>
		/// Gets a value indicating whether the argument is collection-based.
		/// </summary>
		/// <value>
		/// <see langword="true" /> if the argument is collection-based; otherwise, 
		/// <see langword="false" />.
		/// </value>
		public bool IsCollection
		{
            get { return typeof(ICollection).IsAssignableFrom(propertyInfo.PropertyType); }
		}

		/// <summary>
		/// Gets a value indicating whether the argument is array-based.
		/// </summary>
		public bool IsArray
		{
            get { return propertyInfo.PropertyType.IsArray; }
		}

		/// <summary>
		/// Gets a value indicating whether the argument is the default argument.
		/// </summary>
		public bool IsDefault
		{
			get { return attribute is DefaultCommandLineArgumentAttribute; }
		}

        /// <summary>
        /// Gets a value indicating whether the argument is the unprocessed argument.
        /// </summary>
        public bool IsUnprocessed
        {
            get { return attribute is UnprocessedCommandLineArgumentAttribute; }
        }

        /// <summary>
        /// Gets a value indicating whether the argument is the command argument.
        /// </summary>
        public bool IsCommand
        {
            get { return attribute is CommandCommandLineArgumentAttribute; }
        }

        #endregion

		#region Public Instance Methods

		/// <summary>
		/// This method pushes the value from this argument into the matching property into the target object.
		/// </summary>
		/// <param name="target">The object on which the value of the argument should be set.</param>
		[ReflectionPermission(SecurityAction.Demand, Flags = ReflectionPermissionFlag.NoFlags)]
		public void SetTargetValue(object target)
		{
			if (IsArray)
			{
				propertyInfo.SetValue(target, values.ToArray(valueType), BindingFlags.Default, null, null, CultureInfo.InvariantCulture);
			}
			else if (IsCollection)
			{
                object value = propertyInfo.GetValue(target,
                    BindingFlags.Default, null, null, CultureInfo.InvariantCulture);

				// If value of property is null, create new instance of collection 
				if (value == null)
				{
					value = Activator.CreateInstance(propertyInfo.PropertyType,
						BindingFlags.Public | BindingFlags.Instance, null, null, CultureInfo.InvariantCulture);

					propertyInfo.SetValue(target, value, BindingFlags.Default,
						null, null, CultureInfo.InvariantCulture);
				}

				try
				{
                    // Invoke the Add method a bunch of times to add the elements
                    foreach (object item in values)
					{
						addMethod.Invoke(value, BindingFlags.Default, null,
							new object[] { item }, CultureInfo.InvariantCulture);
					}
				}
				catch (Exception ex)
				{
					// So that the Add method can throw parsing errors
					if (ex.InnerException is CommandLineArgumentException)
						throw ex.InnerException;
					else
						throw new NotSupportedException(String.Format(CultureInfo.InvariantCulture,
							CommandLineParserResources.ProblemInvokingAddMethod, Name), ex);
				}
			}
			else
			{
				if (values.Count > 0)
				{
					propertyInfo.SetValue(target, values[0],
						BindingFlags.Default, null, null, CultureInfo.InvariantCulture);
				}
			}
		}

        /// <summary>
        /// This method pulls the values from the matching property in the target into this argument.  Existing values are
        /// overwritten.
        /// </summary>
        /// <param name="target">The object from which the value of the argument should be set.</param>
        [ReflectionPermission(SecurityAction.Demand, Flags = ReflectionPermissionFlag.NoFlags)]
        public void GetTargetValue(object target)
        {
            values.Clear();

            object value = propertyInfo.GetValue(target, BindingFlags.Default, null, null, CultureInfo.InvariantCulture);
            
            if (value == null)
                return;
        
            if (IsArray)
            {
                values.AddRange((ICollection)value);
            }
            else if (IsCollection)
            {
                IEnumerator enumerator = (IEnumerator)getEnumeratorMethod.Invoke(
                    value, BindingFlags.Default, null, null, CultureInfo.InvariantCulture);
                
                while (enumerator.MoveNext())
                {
                    values.Add(enumerator.Current);
                }
            }
            else
            {
                values.Add(value);
            }
        }

        /// <summary>
		/// Assigns the specified value to the argument.
		/// </summary>
		/// <param name="value">The value that should be assigned to the argument.</param>
		/// <exception cref="CommandLineArgumentException">
		/// <para>Duplicate argument OR invalid value.</para>
		/// </exception>
		public void SetArgumentValue(string value)
		{
			if (HasArgumentValue && !AllowMultiple)
			{
				throw new CommandLineArgumentException(String.Format(CultureInfo.InvariantCulture,
					CommandLineParserResources.DuplicateCommandLineArgument, Name));
			}

			// Don't do any processing if this is to be an unprocessed argument list (which by definition must 
			// be an array or a collection)
			if (IsUnprocessed)
			{
				values.Add(value);
				return;
			}
			
    		// Specifically not initialized so we can see if the big if block below has any holes
    		object newValue = null;

            // Null is only valid for bool variables.  Empty string is never valid
            if ((value == null && ValueType != typeof(bool)) || (value != null && value.Length == 0))
            {
                throw new CommandLineArgumentException(String.Format(
                    CultureInfo.InvariantCulture, CommandLineParserResources.InvalidValueForCommandLineArgument, value, Name));
            }
            
            try
            {
                if (ValueType == typeof(string))
                {
                    newValue = value;
                }
                else if (ValueType == typeof(bool))
                {
                    if (value == null || value == "+")
                    {
                        newValue = (object)true;
                    }
                    else if (value == "-")
                    {
                        newValue = (object)false;
                    }
                }
                else
                {
                    if (ValueType.IsEnum)
                    {
                        try
                        {
                            newValue = Enum.Parse(ValueType, value, true);
                        }
                        catch (ArgumentException ex)
                        {
                            string message = CommandLineParserResources.InvalidValueForCommandLineArgumentWithValid;
                            
                            foreach (object obj in Enum.GetValues(ValueType))
                            {
                                message += obj.ToString() + ", ";
                            }
                            
                            // strip last ,
                            message = message.Substring(0, message.Length - 2) + ".";
                            
                            throw new CommandLineArgumentException(String.Format(CultureInfo.InvariantCulture, message, value, Name), ex);
                        }
                    }
                    else
                    {
                        // If there is an initializer class... 
                        if (this.attribute.Initializer != null)
                        {
                            // Look for a public static Parse method on the initializer class
                            System.Reflection.MethodInfo parseMethod = attribute.Initializer.GetMethod(
                                "Parse", BindingFlags.Public | BindingFlags.Static, null,
                                CallingConventions.Standard, new Type[] { typeof(string) }, null);

                            if (parseMethod != null)
                            {
                                // Call the Parse method
                                newValue = parseMethod.Invoke(null, BindingFlags.Default, null,
                                    new object[] { value }, CultureInfo.InvariantCulture);
                            }
                            else
                            {
                                throw new CommandLineArgumentException(String.Format(CultureInfo.InvariantCulture,
                                    CommandLineParserResources.InvalidInitializerClassForCommandLineArgument,
                                    attribute.Initializer.ToString(), Name));
                            }
                        }
                        else
                        {
                            // Look for a public static Parse method on the type of the property
                            System.Reflection.MethodInfo parseMethod = ValueType.GetMethod(
                                "Parse", BindingFlags.Public | BindingFlags.Static, null,
                                CallingConventions.Standard, new Type[] { typeof(string) }, null);

                            if (parseMethod != null)
                            {
                                // Call the Parse method
                                newValue = parseMethod.Invoke(null, BindingFlags.Default, null,
                                    new object[] { value}, CultureInfo.InvariantCulture);
                            }
                            else if (ValueType.IsClass)
                            {
                                // Search for a constructor that takes a string argument
                                ConstructorInfo stringArgumentConstructor =
                                    ValueType.GetConstructor(new Type[] { typeof(string) });

                                if (stringArgumentConstructor != null)
                                {
                                    newValue = stringArgumentConstructor.Invoke(BindingFlags.Default, null, new object[] { value }, CultureInfo.InvariantCulture);
                                }
                            }
                            else
                            {
                                throw new CommandLineArgumentException(String.Format(
                                    CultureInfo.InvariantCulture, CommandLineParserResources.NoWayToInitializeTypeFromString, Name));
                            }
                        }
                    }

                    // TODO-johnls-6/5/2007: This should go after each Invoked call
                    if (newValue == null)
                    {
                        throw new CommandLineArgumentException(String.Format(
                            CultureInfo.InvariantCulture, CommandLineParserResources.UnableToParseValueForArgument, value, Name));
                    }
                }
            }
            catch (Exception e)
            {
                if (!(e is CommandLineArgumentException))
                {
                    throw new CommandLineArgumentException(String.Format(
                        CultureInfo.InvariantCulture, CommandLineParserResources.InvalidValueForCommandLineArgument, value, Name), e);
                }
            }
                
			values.Add(newValue);
		}

        /// <summary>
        /// Returns the value of the argument as a string
        /// </summary>
        public override string ToString()
        {
            if (values.Count == 0)
            {
                return String.Empty;
            }
            else if (this.IsDefault || this.IsUnprocessed || this.IsCommand)
            {
                return String.Join(" ", (string[])values.ToArray(typeof(string))); 
            }
            else
            {
                StringBuilder sb = new StringBuilder();
                IEnumerator enumerator = values.GetEnumerator();
                bool more = enumerator.MoveNext();
                
                while (more)
                {
                    object value = enumerator.Current;
                    
                    more = enumerator.MoveNext();
                
                    if (ValueType == typeof(bool))
                    {
                        sb.AppendFormat(CultureInfo.InvariantCulture, "/{0}{1}{2}", Name, (bool)value ? "+" : "-", more ? " " : "");
                    }
                    else
                    {
                        sb.AppendFormat(CultureInfo.InvariantCulture, "/{0}:{1}{2}", Name, value, more ? " " : "");
                    }
                }
                
                return sb.ToString();
            }
        }

		/// <summary>
		/// Tests if the given command is valid for this argument.  A null or empty command is valid for all arguments.
		/// </summary>
		/// <param name="command"></param>
		/// <returns></returns>
		public bool IsValidCommand(string command)
		{
			if (String.IsNullOrEmpty(command))
			    return true;
			
			string[] commands = attribute.Commands.Split(',', ';');
			
			for (int i = 0; i < commands.Length; i++)
				commands[i] = commands[i].Trim().ToLower();
				
			return Array.Exists<string>(commands, delegate(string item) { return String.CompareOrdinal(item, command) == 0; });
		}

		#endregion

		#region Private Instance Fields

		private ArrayList values;
		private PropertyInfo propertyInfo;
		private MethodInfo addMethod;
		private MethodInfo getEnumeratorMethod;
        private Type valueType;
        private CommandLineArgumentAttribute attribute;

		#endregion
	}

    /// <summary>
    /// Command line parser.
    /// </summary>
    public class CommandLineParser 
    {
        #region Public Instance Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineParser" /> class.  See <see cref="CommandLineParser(Type, Type, CommandLineParserFlags)"/> 
		/// for full description of arguments.
		/// </summary>
		/// <param name="argumentSpecificationType">The <see cref="Type" /> in which the command line arguments are defined.</param>
		/// <exception cref="ArgumentNullException"><paramref name="argumentSpecificationType" /> is a null reference.</exception>
		public CommandLineParser(Type argumentSpecificationType)
			: this(argumentSpecificationType, null, CommandLineParserFlags.None)
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineParser" /> class.  See <see cref="CommandLineParser(Type, Type, CommandLineParserFlags)"/> 
		/// for full description of arguments.
		/// </summary>
		/// <param name="argumentSpecificationType">The <see cref="Type" /> from which the possible command line arguments should be retrieved.</param>
		/// <param name="flags">See <see cref="CommandLineParserFlags"/>.</param>
		/// <exception cref="ArgumentNullException"><paramref name="argumentSpecificationType" /> is a null reference.</exception>
		public CommandLineParser(Type argumentSpecificationType, CommandLineParserFlags flags) : 
			this(argumentSpecificationType, null, flags) 
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineParser" /> class.  See <see cref="CommandLineParser(Type, Type, CommandLineParserFlags)"/> 
		/// for full description of arguments.
        /// </summary>
        /// <param name="argumentSpecificationType">The <see cref="Type" /> in which the command line arguments are defined.</param>
        /// <param name="resourceReaderType">A resource reader object with a <code>GetString</code> method</param>
        /// <exception cref="ArgumentNullException"><paramref name="argumentSpecificationType" /> is a null reference.</exception>
        public CommandLineParser(Type argumentSpecificationType, Type resourceReaderType) :
			this(argumentSpecificationType, resourceReaderType, CommandLineParserFlags.None)
        {
        } 
		
		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineParser" /> class.  Provides a resource reader for localization of the command 
		/// line arguments.  String specified in the command line attributes are used to look up corresponding localized strings in the resources.
		/// Also provides flags to control parsing.
        /// </summary>
		/// <param name="argumentSpecificationType">The <see cref="Type" />  in which the command line arguments are defined.</param>
		/// <param name="resourceReaderType">A resource reader object with a <code>GetString</code> method</param>
		/// <param name="flags">See <see cref="CommandLineParserFlags"/>.</param>
        /// <exception cref="ArgumentNullException"><paramref name="argumentSpecificationType" /> is a null reference.</exception>
        public CommandLineParser(Type argumentSpecificationType, Type resourceReaderType, CommandLineParserFlags flags) 
        {
            if (argumentSpecificationType == null) 
            {
                throw new ArgumentNullException("argumentSpecificationType");
            }

			this.argumentSpecificationType = argumentSpecificationType;
			this.resourceReaderType = resourceReaderType;
			this.flags= flags;

            argumentCollection = new List<CommandLineArgument>();

            foreach (PropertyInfo propertyInfo in argumentSpecificationType.GetProperties(BindingFlags.Instance | BindingFlags.Public)) 
			{
            	object[] attributes = propertyInfo.GetCustomAttributes(typeof(CommandLineArgumentAttribute), false);
				CommandLineArgumentAttribute attribute;

				if (attributes.Length == 1)
					attribute = (CommandLineArgumentAttribute)attributes[0];
				else
					continue;

				if (!(propertyInfo.CanWrite && propertyInfo.CanRead))
					throw new ArgumentException(String.Format(
					    CultureInfo.InvariantCulture, CommandLineParserResources.PropertyShouldBeReadableAndWriteable, propertyInfo.Name));

				attribute.ValueHint = ExternalGetString(attribute.ValueHint);

				if (attribute is DefaultCommandLineArgumentAttribute) 
				{
					if (HasDefaultArgument)
						throw new ArgumentException(CommandLineParserResources.DefaultArgumentAlreadyDefined);
					
					defaultArgument = new CommandLineArgument(attribute, propertyInfo);
				}
				else if (attribute is UnprocessedCommandLineArgumentAttribute)
				{
					if (HasUnprocessedArgument)
						throw new ArgumentException(CommandLineParserResources.UnprocessedArgumentAlreadyDefined);
						
				    unprocessedArgument = new CommandLineArgument(attribute, propertyInfo);
				    
				    if (!unprocessedArgument.AllowMultiple)
				        throw new ArgumentException(CommandLineParserResources.UnprocessedArgumentMustBeArrayOrCollection);
				}
				else if (attribute is CommandCommandLineArgumentAttribute)
				{
					if (HasCommandArgument)
						throw new ArgumentException(CommandLineParserResources.CommandArgumentAlreadyDefined);
						
					commandArgument = new CommandLineArgument(attribute, propertyInfo);
				} 
				else
				{
					attribute.Description = ExternalGetString(attribute.Description);
					
					if (attribute.Description == null)
					{
						throw new ArgumentException(String.Format(
						    CommandLineParserResources.PropertyDoesNotHaveAValueForDescription, propertyInfo.Name));
					}
					
					// If not being case sensitive, make everything lower case.
					if (!CaseSensitive)
					{
						attribute.Name = attribute.Name.ToLower();
					
						if (attribute.ShortName != null)
							attribute.ShortName = attribute.ShortName.ToLower();
					}
				
					argumentCollection.Add(new CommandLineArgument(attribute, propertyInfo));
				}
            }
            
            if (HasUnprocessedArgument && !HasDefaultArgument)
            {
                throw new ArgumentException(CommandLineParserResources.UnprocessedRequiresDefaultArguments);
            }
        }
        
        #endregion

        #region Public Instance Properties

        /// <summary>
        /// Gets a logo banner for the program.
        /// </summary>
        public virtual string LogoBanner 
        {
            get 
            {
				StringBuilder sb = new StringBuilder();
				StringWriter wr = new StringWriter(sb);

				wr.Write(this.Description);
				wr.Write(". ");
				if (this.Configuration != String.Empty)
				{
					wr.Write(this.Configuration);
					wr.Write(" ");
				}
				wr.Write("Version ");
				wr.WriteLine(this.Version);
				wr.WriteLine(this.Copyright);
				wr.WriteLine();
				wr.Flush();
				
				return sb.ToString();
			}
        }

        /// <summary>
        /// A name that will be used in the <see cref="Usage"/> property.  Defaults to <see cref="ProcessName"/> and can
        /// be overridden.
        /// </summary>
        public string CommandName
        {
            get
            {
                if (commandName == null)
                    commandName = ProcessName;
                
                return commandName;
            }
            set
            {
                commandName = value;
            }
        }

        /// <summary>
        /// Gets a suggest name for the MSBuild task, which by default is the name of the argument specification class.
        /// </summary>
        public string TaskName
        {
            get
            {
				if (taskName == null)
				{
					string fullTypeName = argumentSpecificationType.ToString();
					int n = fullTypeName.LastIndexOfAny(new char[] {'.', '+'});
	                
					taskName = fullTypeName.Substring(n == -1 ? 0 : n + 1).ToLower();
				}
				
				return taskName;
            }
        }

        /// <summary>
        /// Gets the name of the current process
        /// </summary>
        public string ProcessName
        {
            get
            {
                return Path.GetFileNameWithoutExtension(Process.GetCurrentProcess().MainModule.FileName).ToLower();
            }
		}

        /// <summary>
        /// Gets the usage instructions.  For command base command lines this is the list of commands and their 
        /// associated descriptions.
        /// </summary>
        /// <value>The usage instructions.</value>
        public virtual string Usage 
        {
            get 
            {
                return GetUsage(null, -1);
            }
        }

        /// <summary>
        /// Gets the total number of processed arguments.
        /// </summary>
        public int ArgumentCount
        {
            get 
            {
				int count = 0;
				
				foreach(CommandLineArgument argument in argumentCollection) 
				{
					if (argument.HasArgumentValue) 
					{
						count += argument.Count;
					}
				}

				if (HasDefaultArgument) 
				{
					count += defaultArgument.Count;
				}
				
				return count;
			}
        }

        /// <summary>
        /// Gets a value indicating if command line processing is case sensitive
        /// </summary>
        public bool CaseSensitive
        {
            get
            {
                return (flags & CommandLineParserFlags.CaseSensitive) != 0;
            }
        }

        /// <summary>
		/// Gets or sets the copyright information to display in the logo. This value is automatically obtained from either 
		/// <see cref="CommandLineCopyrightAttribute"/> or <see cref="AssemblyCopyrightAttribute"/> that is specified for the assembly 
		/// containing the argument specification type. 
		/// </summary>
		public string Copyright
		{
		    get
		    {
				if (copyright == null)
				{
					object[] copyrightAttributes = argumentSpecificationType.GetCustomAttributes(typeof(CommandLineCopyrightAttribute), false);
					
					if (copyrightAttributes.Length > 0)
					{
						CommandLineCopyrightAttribute copyrightAttribute = (CommandLineCopyrightAttribute)copyrightAttributes[0];

						// This one is localizable
						copyright = ExternalGetString(copyrightAttribute.Copyright);
					}
					else
					{
						copyrightAttributes = argumentSpecificationType.Assembly.GetCustomAttributes(typeof(AssemblyCopyrightAttribute), false);

						if (copyrightAttributes.Length > 0)
						{
							AssemblyCopyrightAttribute copyrightAttribute = (AssemblyCopyrightAttribute)copyrightAttributes[0];

							// This one is just used literally
							copyright = copyrightAttribute.Copyright;
						}
						else
						{
							throw new ArgumentException(CommandLineParserResources.CopyrightAttributesNotFound);
						}
					}
				}

				return copyright.Replace("\xA9", "(c)");
			}
		    
		    set 
		    {
		        copyright = value;
		    }
		}

		/// <summary>
		/// Gets or sets the configuration string to display in the logo.  Usually set to Debug for debug builds and
		/// left empty for release builds.  This value is automatically obtained from the <see cref="AssemblyConfigurationAttribute"/> 
		/// that is specified for the assembly containing the argument specification type.
		/// </summary> 
		public string Configuration
		{
			get
			{
				if (configuration == null)
				{
					object[] configurationAttributes = argumentSpecificationType.GetCustomAttributes(typeof(CommandLineConfigurationAttribute), false);

					if (configurationAttributes.Length > 0)
					{
						CommandLineConfigurationAttribute configurationAttribute = (CommandLineConfigurationAttribute)configurationAttributes[0];

						// This is localizable
						configuration = ExternalGetString(configurationAttribute.Configuration);
					}
					else
					{
						configurationAttributes = argumentSpecificationType.Assembly.GetCustomAttributes(typeof(AssemblyConfigurationAttribute), false);
						
						if (configurationAttributes.Length > 0)
						{
							AssemblyConfigurationAttribute configurationAttribute = (AssemblyConfigurationAttribute)configurationAttributes[0];

							// This is literal
							configuration = configurationAttribute.Configuration;
						}
						else
						{
							configuration = String.Empty;
						}
					}
				}

				return configuration;
			}

			set
			{
				configuration = value;
			}
		}

		/// <summary>
		/// Gets or sets the version string to display in the logo.  Set by default to the version of the entry assembly or 
		/// executing assembly if the entry assembly has no version. This value is automatically obtained from the <see cref="AssemblyFileVersionAttribute"/> 
		/// that is specified for the assembly containing the argument specification type.
		/// </summary>
		public string Version
		{
			get
			{
				if (version == null)
				{
					Assembly assembly = argumentSpecificationType.Assembly;
					object[] versionAttributes = assembly.GetCustomAttributes(typeof(AssemblyFileVersionAttribute), false);

					if (versionAttributes.Length > 0)
					{
						AssemblyFileVersionAttribute versionAttribute = (AssemblyFileVersionAttribute)versionAttributes[0];

						version = versionAttribute.Version;
					}
					else
					{
                        throw new ArgumentException(CommandLineParserResources.VersionAttributesNotFound);
					}
				}
				
				return version;
			}

			set
			{
				version = value;
			}
		}

		/// <summary>
		/// Gets or sets the version string to display in the logo.  Set by default to the version of the entry assembly or 
		/// executing assembly if the entry assembly has no version. This value is automatically obtained from the <see cref="CommandLineDescriptionAttribute"/>
		/// if it exists in the assembly containing the argument specification type.  If it does not then <see cref="AssemblyDescriptionAttribute"/> is used
		/// from the same assembly.
		/// </summary>
		public string Description
		{
			get
			{
				if (description == null)
				{
					object[] descriptionAttributes = argumentSpecificationType.GetCustomAttributes(typeof(CommandLineDescriptionAttribute), false);

					if (descriptionAttributes.Length > 0)
					{
						CommandLineDescriptionAttribute descriptionAttribute = (CommandLineDescriptionAttribute)descriptionAttributes[0];

						// This value can be localized
						description = ExternalGetString(descriptionAttribute.Description);
					}
					else
					{
						descriptionAttributes = argumentSpecificationType.Assembly.GetCustomAttributes(typeof(AssemblyDescriptionAttribute), false);

						if (descriptionAttributes.Length > 0)
						{
							AssemblyDescriptionAttribute descriptionAttribute = (AssemblyDescriptionAttribute)descriptionAttributes[0];

							// Use this value literally
							description = descriptionAttribute.Description;
						}
						else
						{
                            throw new ArgumentException(CommandLineParserResources.DescriptionAttributesNotFound);
						}
					}
				}
				
				return description;
			}
			
			set 
			{
				description = value;
			}
		}

        /// <summary>
        /// Gets the command line arguments in cannonical form
        /// </summary>
        public string Arguments
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                List<CommandLineArgument>.Enumerator enumerator = argumentCollection.GetEnumerator();

                if (HasCommandArgument)
                {
                    sb.Append(commandArgument.ToString());
                }

                while (enumerator.MoveNext())
                {
                    CommandLineArgument argument = enumerator.Current;
                    string s = argument.ToString();

                    sb.AppendFormat("{0}{1}", (sb.Length > 0 && s.Length > 0) ? " " : "", s);
                }
                
                if (HasDefaultArgument)
                {
                    string s = defaultArgument.ToString();
                    
                    if (s.Length > 0)
                        sb.Append(" " + s);
                }

                if (HasUnprocessedArgument)
                {
                    string s = unprocessedArgument.ToString();
                    
                    if (s.Length > 0)
                        sb.Append(" " + s);
                }

                return sb.ToString();
            }
        }
        
		/// <summary>
		/// Gets a boolean indicating if the parser has a default command line argument.
		/// </summary>
		public bool HasDefaultArgument
		{
			get
			{
				return defaultArgument != null;
			}
		}

		/// <summary>
		/// Gets a boolean indicating if the parser has an unprocessed command line argument.
		/// </summary>
		public bool HasUnprocessedArgument
		{
			get
			{
				return unprocessedArgument != null;
			}
		}

		/// <summary>
		/// Gets a flag indicating if the parser has a a command argument.
		/// </summary>
		public bool HasCommandArgument
		{
			get
			{
				return commandArgument != null;
			}
		}

        #endregion

        #region Public Instance Methods

        /// <summary>
        /// Parse command line arguments and set a target type instance with argument values.
        /// </summary>
        /// <param name="args"></param>
        /// <param name="target"></param>
        public void ParseAndSetTarget(string[] args, object target)
        {
            Parse(args);
            SetTarget(target);
        }

        /// <summary>
        /// Parses an argument list.
        /// </summary>
        /// <param name="args">The arguments to parse.</param>
        /// <exception cref="ArgumentNullException"><paramref name="target" /> is a null reference.</exception>
        /// <exception cref="ArgumentException">The <see cref="Type" /> of <paramref name="target" /> does not match the argument specification that was used to initialize the parser.</exception>
        public void Parse(string[] args) 
        {
            if (args == null || args.Length == 0)
                return;

            foreach (string argument in args)
            {
                if (argument.Length == 0)
                    continue;

                if (HasDefaultArgument && defaultArgument.HasArgumentValue && HasUnprocessedArgument)
                {
                    // We have a default argument now, everything else should be left unprocessed
                    unprocessedArgument.SetArgumentValue(argument);
                    continue;
                }

                switch (argument[0])
                {
                    case '-':
                    case '/':
                        int endIndex = argument.IndexOfAny(new char[] { ':', '+', '-' }, 1);
                        string argumentName = argument.Substring(1, endIndex == -1 ? argument.Length - 1 : endIndex - 1);
                        string argumentValue;

                        if (argumentName.Length + 1 == argument.Length)
                        {
                            // There's nothing left for a value
                            argumentValue = null;
                        }
                        else if (argument.Length > 1 + argumentName.Length && argument[1 + argumentName.Length] == ':')
                        {
                            argumentValue = argument.Substring(argumentName.Length + 2);
                        }
                        else
                        {
                            argumentValue = argument.Substring(argumentName.Length + 1);
                        }

                        if (!CaseSensitive)
                        {
                            argumentName = argumentName.ToLower();
                        }

                        CommandLineArgument arg;

                        // By this point the strings have either been made the same case or not, and 
                        // we only care about equality, so an ordinal comparison is fine.
                        if (HasCommandArgument)
                        {
                            string command;
                            
                            // If the cammand has not yet been set then the command string is assumed to be default
                            // or empty command, otherwise it is whatever the command value is.
                            if (HasCommandArgument)
                                command = commandArgument.ToString().ToLower();
                            else
                                command = String.Empty;
                        
                            arg = argumentCollection.Find(
                                delegate(CommandLineArgument item)
                                {
                                    return
                                        item.IsValidCommand(command) &&
                                        String.CompareOrdinal(item.Name, argumentName) == 0 ||
                                        String.CompareOrdinal(item.ShortName, argumentName) == 0;
                                });
                        }
                        else
                        {
                            arg = argumentCollection.Find(
                                delegate(CommandLineArgument item)
                                {
                                    return
                                        String.CompareOrdinal(item.Name, argumentName) == 0 ||
                                        String.CompareOrdinal(item.ShortName, argumentName) == 0;
                                });
                        }

                        if (arg == null)
                        {
                            throw new CommandLineArgumentException(
                                String.Format(CultureInfo.InvariantCulture, CommandLineParserResources.UnknownArgument, argument));
                        }
                        else
                        {
                            arg.SetArgumentValue(argumentValue);
                        }
                        break;
                    default:
                        if (HasCommandArgument && !commandArgument.HasArgumentValue)
                        {
                            string command = argument.ToLower();
                            
                            if (!commandArgument.IsValidCommand(command))
                            {
                                throw new CommandLineArgumentException(
                                    String.Format(CultureInfo.InvariantCulture, CommandLineParserResources.UnknownCommandArgument, command));
                            }
                            
                            commandArgument.SetArgumentValue(command);
                        }
                        else if (HasDefaultArgument)
                        {
                            defaultArgument.SetArgumentValue(argument);
                        }
                        else
                        {
                            throw new CommandLineArgumentException(
                                String.Format(CultureInfo.InvariantCulture, CommandLineParserResources.UnknownArgument, argument));
                        }
                        break;
                }
            }
        }
        
        /// <summary>
        /// Initialize a target type instance from parsed arguments.
        /// </summary>
        /// <param name="target">The target type to initialize</param>
        public void SetTarget(object target)
        {
            // NOTE: We typically do this two part parsing because of arrays.  If the target type is an array, there is no efficient 
            // way to keep adding entries to the array as arguments are seen.  Hence we gather them all up and blast them in at once.
            // But it also useful for MSBuild tasks where you don't have a textual command line at all, and for serializing command 
            // line arguments in a standard way.

            if (target == null)
            {
                throw new ArgumentNullException("target");
            }

            // Ensure that the target type is the same as the specification type or derived from it
            if (!argumentSpecificationType.IsAssignableFrom(target.GetType()))
            {
                throw new ArgumentException(String.Format(
                    CultureInfo.InvariantCulture, CommandLineParserResources.TypeNotDerivedFromType,
                    target.GetType().ToString(), argumentSpecificationType.GetType().ToString()));
            }

            foreach (CommandLineArgument arg in argumentCollection)
            {
                arg.SetTargetValue(target);
            }

            // These arguments are not part of the main collection
            if (HasCommandArgument)
            {
                commandArgument.SetTargetValue(target);
            }

            if (HasDefaultArgument)
            {
                defaultArgument.SetTargetValue(target);
            }

            if (HasUnprocessedArgument)
            {
                unprocessedArgument.SetTargetValue(target);
            }
        }

        /// <summary>
        /// Initialize parsed arguments from a target type instance.
        /// </summary>
        /// <param name="target">The target type to initialize</param>
        public void GetTarget(object target)
        {
            if (target == null)
            {
                throw new ArgumentNullException("target");
            }

            // Ensure that the target type is the same as the specification type or derived from it
            if (!argumentSpecificationType.IsAssignableFrom(target.GetType()))
            {
                throw new ArgumentException(String.Format(
                    CultureInfo.InvariantCulture, CommandLineParserResources.TypeNotDerivedFromType,
                    target.GetType().ToString(), argumentSpecificationType.GetType().ToString()));
            }

            foreach (CommandLineArgument arg in argumentCollection)
            {
                arg.GetTargetValue(target);
            }

            // These two arguments are not part of the main collection
            if (HasDefaultArgument)
            {
                defaultArgument.GetTargetValue(target);
            }

            if (HasUnprocessedArgument)
            {
                unprocessedArgument.GetTargetValue(target);
            }
        }

        /// <summary>
        /// Gets the usage string for a specific command.
        /// </summary>
        /// <param name="command">The command for which to obtain usage.  Pass <value>null</value> to obtain default usage.</param>
        /// <returns>The usage string formatted for the default number of display columns.</returns>
        public string GetUsage(string command)
        {
			return GetUsage(command, -1);
        }

        /// <summary>
        /// Gets the usage string for a specific command. 
        /// </summary>
        /// <param name="command">The command for which to obtain usage.  Pass <value>null</value> to obtain default usage.</param>
        /// <param name="lineLength">
		/// Number of columns to format the output for.  Pass <value>-1</value> to format for the active console.  
		/// If there is no console window the default will be 79.  Any line length below 39 is rounded up to 39.
		/// </param>
        /// <returns>The usage string formatted for the specified number of display columns.</returns>
        public string GetUsage(string command, int lineLength)
        {
            StringBuilder helpText = new StringBuilder();
            
            // The empty string is the 'default' command.  This allows some arguments to be valid even if there is no command.  
            if (command == null)
				command = String.Empty;

			command = command.Trim();

			if (lineLength == -1)
			{
				lineLength = 79;

				// There may not be a console window.  Sadly, there is no way to check if there is or not!
				try
				{
					lineLength = Console.BufferWidth - 1;
				}
				catch (IOException) { }
			}
			else
			{
				lineLength = Math.Max(lineLength, 39);
			}

			if (HasCommandArgument && !commandArgument.IsValidCommand(command))
			{
				throw new CommandLineArgumentException(CommandLineParserResources.UnknownCommandArgument);
			}

			GetSummaryUsageLine(command, helpText);

			helpText.Append(System.Environment.NewLine);
            helpText.Append(System.Environment.NewLine);

			if (HasCommandArgument && command.Length == 0)
				GetCommandUsageLines(lineLength, helpText);
			else
				GetOptionsUsageLines(command, lineLength, helpText);

            helpText.Append(System.Environment.NewLine);

            return helpText.ToString();
        }

        #endregion

        #region Private Instance Methods
        
        private void GetSummaryUsageLine(string command, StringBuilder helpText)
        {
			helpText.Append(CommandLineParserResources.Usage_Propercase + ": " + this.CommandName.ToLower());

			if (HasCommandArgument)
			{
				if (String.IsNullOrEmpty(command))
					helpText.Append(" <" + CommandLineParserResources.Command_Lowercase + ">");
				else
					helpText.Append(" " + command);
			}

			if (HasOptions(command))
			{
				helpText.Append(" [" + CommandLineParserResources.Options_Lowercase + "]");
			}

			if (HasDefaultArgument)
			{
				string hint;

				if (!String.IsNullOrEmpty(defaultArgument.ValueHint))
					hint = defaultArgument.ValueHint;
				else
					hint = "<" + defaultArgument.Name.ToLower() + ">";

				helpText.Append(" " + hint);

				// We can only have unprocessed arguments with a default argument
				if (unprocessedArgument == null)
				{
					if (defaultArgument.AllowMultiple)
					{
						// Add additional default arguments
						helpText.Append(" [");
						helpText.Append(hint);
						helpText.Append(" ...]");
					}
				}
				else
				{
					// Add the unprocessed arguments hint
					if (!String.IsNullOrEmpty(unprocessedArgument.ValueHint))
						hint = unprocessedArgument.ValueHint;
					else
						hint = "<" + unprocessedArgument.Name.ToLower() + ">";

					helpText.Append(" [");
					helpText.Append(hint);
					helpText.Append(" ...]");
				}
			}
		}

		private void GetCommandUsageLines(int lineLength, StringBuilder helpText)
		{
			object[] attributes = this.argumentSpecificationType.GetCustomAttributes(typeof(CommandLineCommandDescriptionAttribute), false);
			List<string> commands = new List<string>(attributes.Length);
			int maxCommandLength = 0;
			
			// Build a list of all the commands
			foreach (CommandLineCommandDescriptionAttribute attribute in attributes)
			{
				string command = attribute.Command + "    ";
			
				commands.Add(command);

				if (command.Length > maxCommandLength)
					maxCommandLength = command.Length;
			}

			string indent = new string(' ', maxCommandLength);

			foreach (CommandLineCommandDescriptionAttribute attribute in attributes)
			{
				string command = attribute.Command;
				string description = attribute.Description;

				string[] descriptionLines = StringUtility.WordBreak(description, lineLength - maxCommandLength);

				for (int i = 0; i < descriptionLines.Length; i++)
				{
					if (i == 0)
						helpText.Append(command + new String(' ', maxCommandLength - command.Length));
					else
						helpText.Append(indent);

					helpText.Append(descriptionLines[i]);
					helpText.Append(System.Environment.NewLine);
				}
			}
		}

        private bool HasOptions(string command)
        {
			Debug.Assert(command != null);

            if (command.Length == 0)
			{
			    return argumentCollection.Count > 0;
            }
			else
			{
			    foreach (CommandLineArgument argument in argumentCollection)
			    {
				    if (argument.IsValidCommand(command))
				        return true;
                }

                return false;
            }
        }

		private void GetOptionsUsageLines(string command, int lineLength, StringBuilder helpText)
		{
			Debug.Assert(command != null);
			
			if (!HasOptions(command))
			    return;
			
			// Add options to help text
			helpText.Append(CommandLineParserResources.Options_Propercase + ": ");
			helpText.Append(System.Environment.NewLine);
			helpText.Append(System.Environment.NewLine);

			// First build an array of all the options with all the trimmings,
			// and the descriptions with any necessary modifications, such as the 
			// short option.
            List<KeyValuePair<CommandLineArgument, string>> options = 
                new List<KeyValuePair<CommandLineArgument, string>>(argumentCollection.Count);
			int maxOptionLength = 0;

			foreach (CommandLineArgument argument in argumentCollection)
			{
				if (!argument.IsValidCommand(command))
				    continue;
				
				string valueText = "";

				if (argument.ValueHint != null)
				{
					if (argument.ValueHint != String.Empty)
						valueText = ":" + argument.ValueHint;
				}
				else if (argument.ValueType == typeof(string))
				{
					valueText = ":<" + CommandLineParserResources.Text_LowerCase + ">";
				}
				else if (argument.ValueType == typeof(bool))
				{
					valueText = "[+|-]";
				}
				else if (
					argument.ValueType == typeof(FileInfo) ||
					argument.ValueType == typeof(ParsedPath))
				{
					valueText = ":<" + CommandLineParserResources.FileName_Lowercase + ">";
				}
				else if (
					argument.ValueType == typeof(int) ||
					argument.ValueType == typeof(uint) ||
					argument.ValueType == typeof(short) ||
					argument.ValueType == typeof(ushort))
				{
					valueText = ":<" + CommandLineParserResources.Number_Lowercase + ">";
				}
				else
				{
					valueText = ":<" + CommandLineParserResources.Value_Lowercase + ">";
				}

				string name = argument.Name;
				string shortName = argument.ShortName;

				if (shortName != null && name.StartsWith(shortName))
				{
					name = name.Insert(shortName.Length, "[") + "]";
				}

				string option = "/" + name + valueText + " ";

				options.Add(new KeyValuePair<CommandLineArgument,string>(argument, option));

				if (option.Length > maxOptionLength)
					maxOptionLength = option.Length;
			}

			string indent = new string(' ', maxOptionLength);

			// Now go through and append each line of the help text
			foreach (KeyValuePair<CommandLineArgument, string> pair in options)
			{
				CommandLineArgument argument = pair.Key;
				string option = pair.Value;
				string description = argument.Description;
				string name = argument.Name;
				string shortName = argument.ShortName;

				if (shortName != null && !name.StartsWith(shortName))
				{
					description += String.Format(CultureInfo.InvariantCulture, " " + CommandLineParserResources.ShortFormat, shortName);
				}

				string[] descriptionLines = StringUtility.WordBreak(description, lineLength - maxOptionLength);

				for (int i = 0; i < descriptionLines.Length; i++)
				{
					if (i == 0)
						helpText.Append(option + new String(' ', maxOptionLength - option.Length));
					else
						helpText.Append(indent);

					helpText.Append(descriptionLines[i]);
					helpText.Append(System.Environment.NewLine);
				}
			}
		}

		[ReflectionPermission(SecurityAction.Demand, Flags = ReflectionPermissionFlag.NoFlags)]
		private string ExternalGetString(string s)
		{
			if (resourceReaderType == null || s == null || s == String.Empty)
				return s;
		
			PropertyInfo propInfo = resourceReaderType.GetProperty(s, BindingFlags.NonPublic | BindingFlags.Static);
			
			if (propInfo == null)
                throw new ArgumentException(
                    String.Format(CultureInfo.InvariantCulture,
                    CommandLineParserResources.ResourceReaderDoesNotContainString, 
                    resourceReaderType.ToString(), s));

			return (string)propInfo.GetValue(null, null);
		}

		#endregion

        #region Private Instance Fields

        private List<CommandLineArgument> argumentCollection; 
        private CommandLineArgument defaultArgument;
		private CommandLineArgument unprocessedArgument;
		private CommandLineArgument commandArgument;
		private Type argumentSpecificationType;
        private Type resourceReaderType;
		private CommandLineParserFlags flags;
        private string copyright;
        private string configuration;
        private string version;
        private string description;
        private string taskName;
        private string commandName;

        #endregion
    }

	/// <summary>
	/// The exception that is thrown when one of the command line arguments provided 
	/// is not valid.
	/// </summary>
	public sealed class CommandLineArgumentException : ArgumentException
	{
		#region Public Instance Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineArgumentException" /> class.
		/// </summary>
		public CommandLineArgumentException()
			: base()
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineArgumentException" /> class
		/// with a descriptive message.
		/// </summary>
		/// <param name="message">A descriptive message to include with the exception.</param>
		public CommandLineArgumentException(string message)
			: base(message)
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineArgumentException" /> class
		/// with a descriptive message and an inner exception.
		/// </summary>
		/// <param name="message">A descriptive message to include with the exception.</param>
		/// <param name="innerException">A nested exception that is the cause of the current exception.</param>
		public CommandLineArgumentException(string message, Exception innerException)
			: base(message, innerException)
		{
		}

		#endregion
	}

	/// <summary>
	/// Defines a general description for the command line tool
	/// </summary>
	[AttributeUsage(AttributeTargets.Class, AllowMultiple = false, Inherited = true)]
	public class CommandLineDescriptionAttribute : Attribute
	{
		#region Instance Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineDescriptionAttribute" />.
		/// </summary>
		/// <param name="description">Description of the command line program</param>
		public CommandLineDescriptionAttribute(string description)
		{
			this.description = description;
		}

		#endregion

		#region Instance Properties
		/// <summary>
		/// Gets or sets the description of the command line tool.
		/// </summary>
		public string Description
		{
			get { return description; }
			set { description = value; }
		}

		#endregion

		#region Instance Fields

		private string description;

		#endregion
	}

    /// <summary>
    /// Defines a general description for the command line tool
    /// </summary>
    [AttributeUsage(AttributeTargets.Class, AllowMultiple = true, Inherited = true)]
    public class CommandLineCommandDescriptionAttribute : Attribute
    {
        #region Instance Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CommandLineCommandDescriptionAttribute" />.
        /// </summary>
        /// <param name="command">Name of the command line program</param>
        public CommandLineCommandDescriptionAttribute(string command)
        {
            this.command = command;
        }

        #endregion

        #region Instance Properties
        /// <summary>
        /// Gets or sets the description of the command line command.
        /// </summary>
        public string Description
        {
            get { return description; }
            set { description = value; }
        }

        /// <summary>
        /// Gets or sets the command name
        /// </summary>
        public string Command
        {
            get { return command; }
            set { command = value; }
        }

        #endregion

        #region Instance Fields

        private string command;
        private string description;

        #endregion
    }

	/// <summary>
	/// Defines a copyright for a command line tool
	/// </summary>
	[AttributeUsage(AttributeTargets.Class, AllowMultiple = false, Inherited = true)]
	public class CommandLineCopyrightAttribute : Attribute
	{
		#region Instance Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineCopyrightAttribute" />.
		/// </summary>
		/// <param name="copyright">Copyright for the program</param>
		public CommandLineCopyrightAttribute(string copyright)
		{
			this.copyright = copyright;
		}

		#endregion

		#region Instance Properties
		/// <summary>
		/// Gets or sets the copyright argument.
		/// </summary>
		/// <value>The program copyright information.</value>
		public string Copyright
		{
			get { return copyright; }
			set { copyright = value; }
		}

		#endregion

		#region Instance Fields

		private string copyright;

		#endregion
	}

	/// <summary>
	/// Defines a build configuration for a command line tool
	/// </summary>
	[AttributeUsage(AttributeTargets.Class, AllowMultiple = false, Inherited = true)]
	public class CommandLineConfigurationAttribute : Attribute
	{
		#region Instance Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineConfigurationAttribute" />.
		/// </summary>
		/// <param name="configuration">Copyright for the program</param>
		public CommandLineConfigurationAttribute(string configuration)
		{
			this.configuration = configuration;
		}

		#endregion

		#region Instance Properties
		/// <summary>
		/// Gets or sets the configuration name.
		/// </summary>
		/// <value>The program configuration name (usually Debug or Release).</value>
		public string Configuration
		{
			get { return configuration; }
			set { configuration = value; }
		}

		#endregion

		#region Instance Fields

		private string configuration;

		#endregion
	}

	/// <summary>
	/// Defines a property as an command line argument 
	/// </summary>
	[AttributeUsage(AttributeTargets.Property, AllowMultiple = false, Inherited = true)]
	public class CommandLineArgumentAttribute : Attribute
	{
		#region Instance Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="CommandLineArgumentAttribute" /> class
		/// with the specified argument type.
		/// </summary>
		/// <param name="name">Specifies the checking to be done on the argument.</param>
		public CommandLineArgumentAttribute(string name)
		{
            if (String.IsNullOrEmpty(name))
                throw new ArgumentNullException("name");
            
            this.name = name;
		}

		#endregion

		#region Instance Properties

		/// <summary>
		/// Gets or sets the long name of the argument.
		/// </summary>
		public string Name
		{
			get { return name; }
			set { name = value; }
		}

		/// <summary>
		/// Gets or sets the short name of the argument.
		/// </summary>
		public string ShortName
		{
			get { return shortName; }
			set { shortName = value; }
		}

		/// <summary>
		/// Gets or sets the description of the argument.
		/// </summary>
		public string Description
		{
			get { return description; }
			set { description = value; }
		}

		/// <summary>
		/// Gets or sets hint text to describe the argument value.    Setting this to an empty string
		/// can be used to disable the default value hint, such as [+|-] for boolean arguments.
		/// </summary>
		public string ValueHint
		{
			get { return valueHint; }
			set { valueHint = value; }
		}

		/// <summary>
		/// Gets or sets the initializer type.
		/// </summary>
		public Type Initializer
		{
			get { return initializerType; }
			set { initializerType = value; }
		}

        /// <summary>
        /// Gets or sets the commands associated with this argument
        /// </summary>
        public string Commands
        {
            get { return commands; }
            set { commands = value; }
        }

		#endregion

		#region Instance Fields

        private string name;
		private string shortName;
		private string description;
		private string valueHint;
		private Type initializerType;
        private string commands;

		#endregion
	}

	/// <summary>
	/// Used to capture arguments that are not switches.
	/// </summary>
	[AttributeUsage(AttributeTargets.Property, AllowMultiple = false, Inherited = true)]
	public sealed class DefaultCommandLineArgumentAttribute : CommandLineArgumentAttribute
	{
		/// <summary>
		/// Initializes a new instance of the <see cref="DefaultCommandLineArgumentAttribute"/>
		/// </summary>
		/// <param name="name">Name of the default attribute</param>
		public DefaultCommandLineArgumentAttribute(string name) : base(name)
		{
		}
	}

	/// <summary>
	/// Used to capture the remainder of a command line after other arguments have been processed.
	/// </summary>
	[AttributeUsage(AttributeTargets.Property, AllowMultiple = false, Inherited = true)]
	public sealed class UnprocessedCommandLineArgumentAttribute : CommandLineArgumentAttribute
	{
		/// <summary>
		/// Initializes a new instance of the <see cref="UnprocessedCommandLineArgumentAttribute" /> 
		/// </summary>
		public UnprocessedCommandLineArgumentAttribute(string name) : base(name)
		{
		}
	}

    /// <summary>
    /// Used to for command base command lines where one of the arguments is is a command. The command defines 
    /// which of the other arguments are valid for that command.
    /// </summary>
    [AttributeUsage(AttributeTargets.Property, AllowMultiple = false, Inherited = true)]
    public sealed class CommandCommandLineArgumentAttribute : CommandLineArgumentAttribute
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="CommandCommandLineArgumentAttribute" /> 
        /// </summary>
        public CommandCommandLineArgumentAttribute(string name)
            : base(name)
        {
        }
    }

	/// <summary>
	/// Flags that control command line parsing
	/// </summary>
	[Flags]
	public enum CommandLineParserFlags
	{
		/// <summary>
		/// All switches are matched case sensitively.
		/// </summary>
		CaseSensitive = 0x0001,
		/// <summary>
		/// Normal parsing.  All switches are matched case insensitive.
		/// </summary>
		None = 0x0000
	}
}
