// Copyright (c) Microsoft Corporation. All rights reserved.

#include <windows.h>
#include <strsafe.h>

#ifdef _DEBUG
#define ODS(m) OutputDebugStringW(L"CodeRunnerHelper.dll: " m)
#else
#define ODS(m) __noop(m)
#endif

#define MAX_MESSAGE_LENGTH 32768
#define COUNT_OF(x) (sizeof(x)/sizeof(x[0]))
#define LENGTH_OF(x) (COUNT_OF(x) - 1)

void ProcessRequest(wchar_t *pRequest, int requestLength, wchar_t *pReply, int maxReplyLength, int *pReplyLength, BOOL *pDisconnect)
{
	LCID lcid = MAKELCID(MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), SORT_DEFAULT);
	wchar_t const setVarCommand[] = L"setvar\r\n";
	wchar_t const getVarCommand[] = L"getvar\r\n";
	wchar_t const setCdCommand[] = L"setcd\r\n";
	wchar_t const getCdCommand[] = L"getcd";
	wchar_t const terminateCommand[] = L"terminate";

	*pReplyLength = 0;
	*pDisconnect = FALSE;

	if (requestLength >= LENGTH_OF(setVarCommand) + 3 &&  // Command must be at least 'setvar\r\nx=y'   
	    CompareString(lcid, 0, pRequest, LENGTH_OF(setVarCommand), setVarCommand, LENGTH_OF(setVarCommand)) == CSTR_EQUAL)
	{
		// Message format:
		// 
		//   setvar
		//   name=value
		//

		wchar_t *pName = pRequest + LENGTH_OF(setVarCommand);
		wchar_t *pValue = pName + 1;
	
		// Subtract two from requestLength below because we need at least the '=' and one character for the value left after scanning
		while ((pValue - pRequest) < requestLength - 2 && *pName != '\0'  && *pValue != L'=' && *pValue != L'\r' && *pValue != L'\n')
			pValue++;

        BOOL ok = TRUE;
			
		if (*pValue == L'=')
		{
			*pValue++ = 0;
		
			ok = SetEnvironmentVariableW(pName, pValue);	
		}

		// TODO-johnls-12/24/2007: Allow setting multiple environment variables at once?
    	
    	*pReplyLength = 1;
    	*pReply = (ok ? L'T' : L'F');
	}
	else if (requestLength >= LENGTH_OF(getVarCommand) + 1 &&  // Command must be at least 'getvar\r\nx'
	    CompareString(lcid, 0, pRequest, LENGTH_OF(getVarCommand), getVarCommand, LENGTH_OF(getVarCommand)) == CSTR_EQUAL)
	{
		// Message format:
		// 
		//   getvar
		//   name
		//

		wchar_t *pName = pRequest + LENGTH_OF(getVarCommand);
	
    	*pReplyLength = GetEnvironmentVariableW(pName, pReply, maxReplyLength);	
    }
	else if (requestLength >= LENGTH_OF(setCdCommand) + 1 && // Command must be at least 'setcd\r\nx'
	    CompareString(lcid, 0, pRequest, LENGTH_OF(setCdCommand), setCdCommand, LENGTH_OF(setCdCommand)) == CSTR_EQUAL)
	{
		// Message format:
		// 
		//   setcd
		//   dir
		//

		wchar_t *pDir = pRequest + LENGTH_OF(setCdCommand);
	
    	BOOL ok = SetCurrentDirectoryW(pDir);	
    	
    	*pReplyLength = 1;
    	*pReply = (ok ? L'T' : L'F');
    }
	else if (requestLength >= LENGTH_OF(getCdCommand) && 
	    CompareString(lcid, 0, pRequest, LENGTH_OF(getCdCommand), getCdCommand, LENGTH_OF(getCdCommand)) == CSTR_EQUAL)
	{
		// Message format:
		// 
		//   getcd
		//

    	*pReplyLength = GetCurrentDirectoryW(maxReplyLength, pReply);	
    }
	else if (requestLength >= LENGTH_OF(terminateCommand) && 
	    CompareString(lcid, 0, pRequest, LENGTH_OF(terminateCommand), terminateCommand, LENGTH_OF(terminateCommand)) == CSTR_EQUAL)
	{
		// Message format:
		//
		//   terminate
		//
		
		*pDisconnect = TRUE;
	}
}

// This method is called from the remotely injected thread
DWORD RealControlRemoteEnvironment(void)
{
	LPTSTR pipeName = L"\\\\.\\pipe\\coderunner$"; 
	HANDLE hPipe = CreateFile(pipeName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);

	if (hPipe == INVALID_HANDLE_VALUE) 
	{
		// If we don't get connected then we just bail, because obviously things aren't well in the outside world
		return 0;
	}

	wchar_t pRequest[MAX_MESSAGE_LENGTH]; 
	wchar_t pReply[MAX_MESSAGE_LENGTH]; 
	DWORD actualBytes; 
	BOOL success; 
	BOOL disconnect = FALSE;
 
	while (!disconnect) 
	{ 
		int attemptedBytes = MAX_MESSAGE_LENGTH * sizeof(wchar_t);

		success = ReadFile(hPipe, pRequest, attemptedBytes, &actualBytes, NULL);

		if (!success || actualBytes == 0 || actualBytes % sizeof(wchar_t) != 0) 
			break; 
		
		int requestLength = (actualBytes / sizeof(wchar_t)) - 1;
		
		// Check for null terminated string
		if (pRequest[requestLength] != 0)
			break;
		
		int replyLength;

		ProcessRequest(pRequest, requestLength, pReply, COUNT_OF(pReply), &replyLength, &disconnect);
		
		if (replyLength > 0)
		{
			attemptedBytes = (replyLength + 1) * sizeof(wchar_t);
			
			success = WriteFile(hPipe, pReply, attemptedBytes, &actualBytes, NULL);
			
			if (!success || actualBytes != attemptedBytes) 
				break;
		}
	}

	// Flush the pipe to allow the client to read the pipe's contents before disconnecting. Then disconnect 
	// the pipe, and close the handle to this pipe instance. 
	FlushFileBuffers(hPipe); 
	DisconnectNamedPipe(hPipe); 
	CloseHandle(hPipe); 

	return 0;
}

#ifndef PIPESERVER

#pragma comment(linker, "/EXPORT:ControlRemoteEnvironment=_ControlRemoteEnvironment@0")
extern "C" DWORD WINAPI ControlRemoteEnvironment(void)
{
	DWORD ret = 0;

	__try
	{
		ret = RealControlRemoteEnvironment();
	}
	__except(EXCEPTION_CONTINUE_EXECUTION)
	{
		// We MUST NOT bring down the process into which we were injected!
	}

	return ret;	
}

#endif