﻿using System;
using System.CodeDom.Compiler;
using System.Collections;
using System.Collections.Specialized;
using System.Configuration;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using Microsoft.CSharp;
using System.Collections.Generic;

namespace Microsoft.Tools.CodeRunner
{
    /// <summary>
    /// Code runner build task
    /// </summary>
    [CommandLineCopyright("CopyrightMicrosoft")]
    [CommandLineTitle("CSharpCodeRunner")]
    [CommandLineDescription("CommandLineDescription")]
    public class Csr : IProcessCommandLine, IProcessEnvironment, IProcessConfiguration
    {
        #region Private Fields
        private bool runningFromCommandLine = false;
        private bool debugMessages = false;
        private bool stackTraces = false;
        private bool searchSystemPath = true;
        private bool allowUnsafeCode = false;
        private CommandLineParser parser = null;
        private static readonly string csrEnvironmentVar = "CSR_CONFIG";
        private ParsedPath scriptPath;
        private ParsedPath temporaryFileDirectory;
        private OutputHelper output;
        
        #endregion

        #region Private Properties
        internal CommandLineParser Parser 
        { 
            get 
            { 
                if (parser == null)
                    parser = new CommandLineParser(typeof(Csr), typeof(CsrResources), CommandLineParserFlags.None);
                
                return parser;
            }
        }

        public OutputHelper Output 
        { 
            get { return this.output; }
        }

        #endregion

        #region Public Properties
        /// <summary>
        /// Show help
        /// </summary>
        [CommandLineArgument("help", Description = "HelpSwitchDescription", ShortName = "?")]
        public bool ShowHelp { get; set; }

        [DefaultCommandLineArgument("script", ValueHint = "ScriptSwitchHint")]
        internal ParsedPath ScriptPath { get { return scriptPath; } set { scriptPath = value; } }

        /// <summary>
        /// Script arguments
        /// </summary>
        [UnprocessedCommandLineArgument("arguments", ValueHint = "ArgumentSwitchHint")]
        public string[] Arguments { get; set; }

        /// <summary>
        /// Display debug messages
        /// </summary>
        public bool DebugMessages { get { return debugMessages; } set { debugMessages = value; } }

        /// <summary>
        /// Display line numbers in error messages
        /// </summary>
        public bool StackTraces { get { return stackTraces; } set { stackTraces = value; } }

        /// <summary>
        /// Search system path for script if path not fully qualified
        /// </summary>
        public bool SearchSystemPath { get { return searchSystemPath; } set { searchSystemPath = value; } }

        /// <summary>
        /// Allow unsafe code to be compiled into script
        /// </summary>
        public bool AllowUnsafeCode { get { return allowUnsafeCode; } set { allowUnsafeCode = value; } }
        
        /// <summary>
        /// Exit code from script
        /// </summary>
        public int ExitCode { get; set; }

        #endregion    
        
        #region Construction
        /// <summary>
        /// Default constructor
        /// </summary>
        public Csr()
        {
        }
        
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="outputter">IOutputter to use for the script</param>
        public Csr(IOutputter outputter)
        {
            this.output = new OutputHelper(outputter, CsrResources.ResourceManager);
        }

        #endregion    
        
        #region Public Methods
        /// <summary>
        /// Execute the task
        /// </summary>
        /// <returns>true if task succeeded, false otherwise</returns>
        public bool Execute()
        {
            if (!runningFromCommandLine)
            {
                // Set the command name to the task name and display the canonical comamnd line
                Parser.CommandName = Parser.TaskName;
                
                string arguments = Parser.Arguments;
                Output.Message(MessageImportance.Low, Parser.CommandName + (arguments.Length == 0 ? "" : " " + Parser.Arguments));
            }

            // If the help option appears anywhere, just show the help
            if (this.ShowHelp)
            {
                Output.Message(Parser.LogoBanner);
                Output.Message(Parser.Usage);
                Output.Message("");
                return true;
            }

            // From this point on assume a bad exit code
            this.ExitCode = 1;

            // At this point, we need a source file (the default argument)
            if (this.ScriptPath == null)
            {
                Output.Error(CsrResources.NoScriptSpecified);
                return false;
            }

            // Before we fully qualify the file record whether it had any full or relative path part
            bool justFilenameGiven = this.ScriptPath.IsFilenameOnly;

            // Fully qualify the source
            this.ScriptPath = this.ScriptPath.MakeFullPath();

            if (this.ScriptPath.File == String.Empty)
            {
                Output.Error(CsrResources.NoEmptyFileName);
                return false;
            }
            else if (this.ScriptPath.HasWildcards)
            {
                Output.Error(CsrResources.NoScriptWildcards);
                return false;
            }

            if (this.ScriptPath.Extension == String.Empty)
                this.ScriptPath = new ParsedPath(this.ScriptPath.RootDirectoryAndFile + ".csr");

            // Set publicly visible script path (in this AppDomain).  This also sets the script.rsp file name.
            ScriptEnvironment.ScriptPath = this.ScriptPath;

            if (this.DebugMessages)
            {
                Output.Message(MessageImportance.Low, CsrResources.ClrInstallPath, ScriptEnvironment.ClrInstallPath);
                Output.Message(MessageImportance.Low, CsrResources.NetFxInstallPath, ScriptEnvironment.NetFxInstallPath);
                Output.Message(MessageImportance.Low, CsrResources.NetFxReferenceAssemblyPath, ScriptEnvironment.NetFxReferenceAssemblyPath);
                Output.Message(MessageImportance.Low, CsrResources.TemporaryFileDirectory, TemporaryFileDirectory);
            }

            // Check that the source exists, and optionally search for it in the PATH
            if (!File.Exists(this.ScriptPath))
            {
                if (this.SearchSystemPath && justFilenameGiven)
                {
                    string[] found = PathUtility.FindFileInPaths(
                        System.Environment.GetEnvironmentVariable("PATH").Split(new char[] { Path.PathSeparator }), this.ScriptPath.FileAndExtension);

                    if (found.Length > 0)
                    {
                        this.ScriptPath = new ParsedPath(found[0]);
                        ScriptEnvironment.ScriptPath = this.ScriptPath;

                        if (this.DebugMessages)
                            Output.Message(MessageImportance.Low, CsrResources.ScriptFoundInPath,
                                this.ScriptPath.FileAndExtension, this.ScriptPath.RootAndDirectory);
                    }
                    else
                    {
                        Output.Error(CsrResources.ScriptNotFoundInDirectoryOrPath,
                            this.ScriptPath.FileAndExtension, this.ScriptPath.RootAndDirectory);
            
                        return false;
                    }
                }
                else
                {
                    Output.Error(CsrResources.ScriptNotFound, this.ScriptPath);
                    return false;
                }
            }

            // Now that we have found the scripts full path, the local response file is correct
            ParsedPath localRspFile = ScriptEnvironment.ScriptRspPath;
            ParsedPath sharedRspFile = ScriptEnvironment.SharedRspPath;
            ParsedPath compilerRspFile = ScriptEnvironment.CompilerRspPath;

            if (this.DebugMessages)
            {
                Output.Message(MessageImportance.Low, CsrResources.LocalResponseFilePath, localRspFile,
                    File.Exists(localRspFile) ? CsrResources.Found : CsrResources.NotFound);
                Output.Message(MessageImportance.Low, CsrResources.SharedResponseFilePath, sharedRspFile,
                    File.Exists(sharedRspFile) ? CsrResources.Found : CsrResources.NotFound);
                Output.Message(MessageImportance.Low, CsrResources.CompilerResponseFilePath, compilerRspFile,
                    File.Exists(compilerRspFile) ? CsrResources.Found : CsrResources.NotFound);
            }

            // Combine all the .rsp's with fully qualified paths to all references
            ParsedPath combinedRspFile = CombineResponseFiles(localRspFile, sharedRspFile, compilerRspFile);

            if (DebugMessages)
                Output.Message(MessageImportance.Low, CsrResources.CombinedResponseFile, combinedRspFile);
            
            try
            {
                IDictionary<string, string> providerOptions = new Dictionary<string, string>();
                
                if (String.IsNullOrEmpty(ScriptEnvironment.ComPlusInstallRoot))
                {
                    // Setting the compiler version option should only be done under Fx v3.5 otherwise the option should be not present
                    if (ScriptEnvironment.NetFxVersion == "v3.5")
                        providerOptions.Add("CompilerVersion", ScriptEnvironment.NetFxVersion);
                }
                
                CodeDomProvider provider = new CSharpCodeProvider(providerOptions);
                CompilerParameters parms = new CompilerParameters();
            
                parms.CompilerOptions = "@\"" + combinedRspFile + "\" ";  // Don't forget space at end...
                parms.GenerateExecutable = true;	// This doesn't mean generate a file on disk, it means generate an .exe not a .dll
                
                if (StackTraces)
                {
                    if (!Directory.Exists(TemporaryFileDirectory))
                        Directory.CreateDirectory(this.TemporaryFileDirectory);
                    
                    parms.OutputAssembly = TemporaryFileDirectory.RootAndDirectory + ScriptPath.File + ".exe";
                    parms.GenerateInMemory = false;
                    parms.IncludeDebugInformation = true;
                    parms.CompilerOptions += "/optimize- ";
                }
                else
                {
                    parms.GenerateInMemory = true;
                    parms.CompilerOptions += "/optimize+ ";
                }

                if (this.AllowUnsafeCode)
                    parms.CompilerOptions += "/unsafe+ ";

                if (DebugMessages)
                    Output.Message(MessageImportance.Low, "Compiling script file '{0}'", this.ScriptPath);
                    
                CompilerResults results = provider.CompileAssemblyFromFile(parms, this.ScriptPath);

                if (!DebugMessages)
                    File.Delete(combinedRspFile);

                if (results.Errors.Count > 0)
                {
                    foreach (CompilerError error in results.Errors)
                    {
                        Console.Error.WriteLine(CsrResources.ErrorLine,
                            error.FileName, error.Line, error.Column, error.ErrorNumber, error.ErrorText);
                    }
                    
                    return false;
                }
                else
                {
                    // Take the compiled assembly and invoke it in this appdomain.
                    object ret = null;

                    AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(OnAssemblyResolve);

                    // Get information about the entry point							
                    MethodInfo mainMethod = results.CompiledAssembly.EntryPoint;
                    ParameterInfo[] mainParams = mainMethod.GetParameters();
                    Type returnType = mainMethod.ReturnType;

                    try
                    {
                        if (returnType == typeof(void))
                        {
                            if (mainParams.Length > 0)
                                mainMethod.Invoke(null, new object[1] { this.Arguments });
                            else
                                mainMethod.Invoke(null, null);
                        }
                        else
                        {
                            if (mainParams.Length > 0)
                                ret = mainMethod.Invoke(null, new object[1] { this.Arguments });
                            else
                                ret = mainMethod.Invoke(null, null);
                        }
                    }
                    catch (Exception ex)  // Catch script errors
                    {
                        // When catching a script error here, the actual script exception will be the inner exception
                        Output.Error(String.Format(CsrResources.ExceptionFromScript,
                            ex.InnerException != null ? ex.InnerException.Message : ex.Message));

                        if (this.StackTraces && ex.InnerException != null)
                            Output.Error(ex.InnerException.StackTrace);

                        return false;
                    }

                    if (ret != null && returnType == typeof(int))
                    {
                        this.ExitCode = (int)ret;
                        
                        return (this.ExitCode == 0);
                    }
                }
            }
            catch (Exception ex)  // Catch compilation exceptions
            {
                string message = CsrResources.ExceptionDuringCompile + ex.Message;

                if (ex.InnerException != null)
                    message += ex.InnerException.Message;

                Output.Error(message);
                return false;
            }

            ExitCode = 0;
            
            return true;
        }

        #endregion    

        #region Internal Static Methods
        internal static Assembly OnAssemblyResolve(object sender, ResolveEventArgs args)
        {
            // Assembly resolver, used to load assemblies specified in the .rsp files or command line which would 
            // not normally be found because they are not in the GAC, csr.exe directory or cached image directory.

            ParsedPath scriptRspFile = ScriptEnvironment.ScriptRspPath;
            ParsedPath sharedRspFile = ScriptEnvironment.SharedRspPath;
            ParsedPath runtimeRspFile = ScriptEnvironment.CompilerRspPath;

            List<ParsedPath> refPaths = new List<ParsedPath>();

            if (File.Exists(scriptRspFile))
            {
                refPaths.AddRange(GrabRspFileReferences(scriptRspFile, null));
            }

            if (File.Exists(sharedRspFile))
            {
                refPaths.AddRange(GrabRspFileReferences(sharedRspFile, null));
            }

            if (File.Exists(runtimeRspFile))
            {
                refPaths.AddRange(GrabRspFileReferences(runtimeRspFile, null));
            }

            string assemblyName = args.Name.Substring(0, args.Name.IndexOf(','));

            // Find the first assembly where the file names match
            foreach (ParsedPath path in refPaths)
            {
                if (CaseInsensitiveComparer.Default.Compare(path.File, assemblyName) == 0)
                    return Assembly.LoadFrom(path);
            }

            return null;
        }

        #endregion

        #region Private Properties
        private ParsedPath TemporaryFileDirectory
        {
            get
            {
                if (temporaryFileDirectory == null)
                {
                    temporaryFileDirectory = new ParsedPath(Path.GetTempPath()).Combine(@"CodeRunner\");
                }

                return temporaryFileDirectory;
            }
        }

        #endregion

        #region Private Methods
        private static string CachifyScriptPath(ParsedPath scriptPath)
        {
            Debug.Assert(scriptPath.IsFull);

            StringBuilder sb = new StringBuilder(scriptPath.RootDirectoryAndFile + ".exe");

            sb.Replace('\\', '_');
            sb.Replace(':', '_');
            sb.Replace(' ', '_');

            return sb.ToString();
        }

        private ParsedPath CombineResponseFiles(ParsedPath scriptRspFile, ParsedPath sharedRspFile, ParsedPath compilerRspFile)
        {
            // This function combines existing responses file into one temporary one
            // but with all the referenced assemblies having fully qualified paths, relative to the 
            // original response files location or in one of the system reference assembly directories.
            // Fully qualified references are not adjusted.

            ParsedPath combinedRspFile = new ParsedPath(Path.GetTempFileName());
            ParsedPath[] rspFilePaths = new ParsedPath[3] 
            {
                scriptRspFile, sharedRspFile, compilerRspFile 
            };

            using (StreamWriter wr = File.CreateText(combinedRspFile))
            {
                foreach (var rspFilePath in rspFilePaths)
                {
                    if (!File.Exists(rspFilePath))
                        continue;

                    List<ParsedPath> refPaths = GrabRspFileReferences(rspFilePath, this.DebugMessages ? this.Output : null);

                    if (refPaths == null)
                        return null;

                    foreach (var refPath in refPaths)
                    {
                        wr.Write("/r:\"");
                        wr.Write(refPath);
                        wr.Write("\"\n");
                    }
                }
            }

            return combinedRspFile;
        }

        private static List<ParsedPath> GrabRspFileReferences(ParsedPath rspFile, OutputHelper output)
        {
            string text;
            StringBuilder sb = new StringBuilder();

            using (StreamReader reader = new StreamReader(rspFile))
            {
                text = reader.ReadToEnd();
            }

            Regex r = new Regex("/r[^:]*:(.*)$", RegexOptions.IgnoreCase | RegexOptions.Multiline);

            Match m = r.Match(text);

            List<ParsedPath> refPaths = new List<ParsedPath>();
            List<ParsedPath> searchPaths = new List<ParsedPath>();
            
            searchPaths.Add(null);
            searchPaths.Add(ScriptEnvironment.NetFxReferenceAssemblyPath.Combine(@"v4.0\"));
            searchPaths.Add(ScriptEnvironment.NetFxReferenceAssemblyPath.Combine(@"v3.5\"));
            searchPaths.Add(ScriptEnvironment.NetFxReferenceAssemblyPath.Combine(@"v3.0\"));
            searchPaths.Add(ScriptEnvironment.ClrInstallPath);

            while (m.Success)
            {
                string s = m.Groups[1].Value;

                // Chop off trailing '\r'; won't be there if last line of file doesn't end in \r\n
                if (s[s.Length - 1] == '\r')
                    s = s.Substring(0, s.Length - 1);

                s = Environment.ExpandEnvironmentVariables(s);

                ParsedPath rawPath = new ParsedPath(s, PathType.File);
                ParsedPath refPath = null;

                searchPaths[0] = new ParsedPath(rspFile.RootAndDirectory);

                foreach (var searchPath in searchPaths)
                {
                    ParsedPath path = rawPath.MakeFullPath(searchPath);
                
                    if (File.Exists(path))
                    {
                        refPath = path;
                        break;
                    }
                }

                if (refPath != null)
                {
                    refPaths.Add(refPath);
                }
                else if (output != null)
                {
                    output.Warning(CsrResources.ReferenceAssemblyNotFound, rawPath, rspFile);
                }

                m = m.NextMatch();
            }

            return refPaths;
        }
        
        private void GetBoolSetting(CsrSection section, string setting, ref bool val)
        {
            try
            {
                for (int i = 0; i < section.Settings.Count; i++)
                {
                    TypeElement element = section.Settings[i];

                    if (String.Compare(element.Key, setting, true) == 0)
                    {
                        if (!TryParseTrueFalse(element.Value, ref val))
                            Output.Warning(CsrResources.ErrorParsingBooleanInConfig, element.Value, element.Key);
                        
                        return;
                    }
                }
            }
            catch (ConfigurationException e)
            {
                Output.Warning(e.Message);
            }
        }

        private void GetBoolSetting(IDictionary dict, string setting, ref bool val)
        {
            string temp = (string)dict[setting];

            if (!String.IsNullOrEmpty(temp))
            {
                if (!TryParseTrueFalse(temp, ref val))
                    Output.Warning(CsrResources.ErrorParsingBooleanInEnvironment, temp, setting, csrEnvironmentVar);
            }
        }
        
        private static bool TryParseTrueFalse(string s, ref bool b)
        {
            if ((String.Compare(s, CsrResources.No, true) == 0 ||
                String.Compare(s, CsrResources.False, true) == 0))
            {
                b = false;
                return true;
            }
            else if ((String.Compare(s, CsrResources.Yes, true) == 0 ||
                String.Compare(s, CsrResources.True, true) == 0))
            {
                b = true;
                return true;
            }
            
            return false;
        }

        #endregion

        #region IProcessCommandLine Members

        bool IProcessCommandLine.ProcessCommandLine(string[] args)
        {
            runningFromCommandLine = true;
            
            CommandLineParser parser = new CommandLineParser(typeof(Csr), typeof(CsrResources));

            try
            {
                parser.ParseAndSetTarget(args, this);
            }
            catch (CommandLineArgumentException e)
            {
                Output.Error(e.Message);
                return false;
            }

            return true;
        }

        #endregion

        #region IProcessEnvironment Members

        bool IProcessEnvironment.ProcessEnvironment()
        {
            string allVars = Environment.GetEnvironmentVariable(csrEnvironmentVar);
            
            if (allVars == null)
                return true;	
                
            string[] vars = allVars.Split(';');

            ListDictionary environ = new ListDictionary();
            
            if (vars.Length > 0)
            {
                foreach (string var in vars)
                {
                    string[] parts = var.Split('=');
                    
                    if (parts.Length == 1)
                    {
                        environ[parts[0].Trim()] = String.Empty;
                    }
                    else if (parts.Length == 2 && parts[0].Length > 0)
                    {
                        environ[parts[0].Trim()] = parts[1].Trim();
                    }
                }
            }
            
            if (environ.Count == 0)
                return true;

            GetBoolSetting(environ, "DebugMessages", ref debugMessages);
            GetBoolSetting(environ, "StackTraces", ref stackTraces);
            GetBoolSetting(environ, "SearchSystemPath", ref searchSystemPath);
            GetBoolSetting(environ, "AllowUnsafeCode", ref allowUnsafeCode);
            
            return true;
        }

        #endregion

        #region IProcessConfig Members

        bool IProcessConfiguration.ProcessConfiguration(ConfigurationUserLevel userLevel)
        {
            try
            {
                Configuration config = ConfigurationManager.OpenExeConfiguration(userLevel);
                CsrSection section = null;
                
                if (!config.HasFile)
                    return true;
                
                try
                {
                    section = config.GetSection("csr") as CsrSection;
                }
                catch (ConfigurationException e)
                {
                    Output.Error("Problem loading program configuration - {0}", e.Message);
                }

                if (section == null)
                {
                    // This means that either there was no section or that it was improperly specified in
                    // the .config file.  Make sure that the configSection is at the top of the file.
                    return true;
                }

                GetBoolSetting(section, "DebugMessages", ref debugMessages);
                GetBoolSetting(section, "StackTraces", ref stackTraces);
                GetBoolSetting(section, "SearchSystemPath", ref searchSystemPath);
                GetBoolSetting(section, "AllowUnsafeCode", ref allowUnsafeCode);
            }
            catch (ConfigurationErrorsException)
            {
                Output.Warning("Application configuration could not be loaded");
                return false;
            }
            
            return true;
        }

        #endregion
    }
}
